import os
import shutil
from tempfile import mkstemp, mkdtemp
import zipfile
import contextlib
import gzip
import tarfile
from ..utils import _uncompress_file


def test_uncompress():
    # Create a zipfile
    dtemp = mkdtemp()
    ztemp = os.path.join(dtemp, 'test.zip')
    ftemp = 'test'
    try:
        with contextlib.closing(zipfile.ZipFile(ztemp, 'w')) as testzip:
            testzip.writestr(ftemp, 'test')
        _uncompress_file(ztemp)
        assert (os.path.exists(os.path.join(dtemp, ftemp)))
        shutil.rmtree(dtemp)

        dtemp = mkdtemp()
        ztemp = os.path.join(dtemp, 'test.tar')
        ftemp = 'test'

        # Create dummy file in the dtemp folder, so that the finally statement
        # can easily remove it
        fd, temp = mkstemp(dir=dtemp)
        os.close(fd)
        with contextlib.closing(tarfile.open(ztemp, 'w')) as tar:
            tar.add(temp, arcname=ftemp)
        _uncompress_file(ztemp)
        assert (os.path.exists(os.path.join(dtemp, ftemp)))
        shutil.rmtree(dtemp)

        dtemp = mkdtemp()
        ttemp = os.path.join(dtemp, 'test')
        with open(ttemp, 'wb') as f:
            f.write('test'.encode('utf-8'))
        ztemp = os.path.join(dtemp, 'test.gz')
        with open(ttemp, 'rb') as f_in, gzip.open(ztemp, 'wb') as f_out:
            shutil.copyfileobj(f_in, f_out)
        _uncompress_file(ztemp)
        # test.gz gets uncompressed into test
        assert (os.path.exists(ttemp))
        shutil.rmtree(dtemp)

        # try to uncompress a corrupted file
        dtemp = mkdtemp()
        ttemp = os.path.join(dtemp, 'test')
        with open(ttemp, 'wb') as fid:
            fid.write('test'.encode('utf-8'))
        try:
            _uncompress_file(ttemp)
        except Exception as e:
            assert isinstance(e, IOError)


    finally:
        # all temp files are created into dtemp except temp
        if os.path.exists(dtemp):
            shutil.rmtree(dtemp)



"""
fetching function to retrieve example dataset, using nilearn
fetching convention.

The parts of the nilearn fetching utils that have an obvious
meaning are directly copied. The rest is annoted.
"""
# -*- coding: utf-8 -*-
import os
import requests
import shutil
import urllib
from collections import namedtuple
import contextlib
import warnings

from ..datasets.utils import md5_hash, _check_if_exists, \
    _uncompress_file, \
    _md5_sum_file, get_data_dir

# in nilearn, urllib is used. Here the request package will be used
# trying to use requests as much as possible (everything except the
# parsing function)

# current data has been pulled from bigml, which require authentication
# for downloading. So for now we download the data from github
# however, the data differ a little bit from the two sources
# so we either have to implement login into _fetch_data
# or to reverse-engineer the processing script that can transform the data
# from git
# to the data from bigml
# this is true for bigml and midwest survey


DatasetInfo = namedtuple('DatasetInfo',
                         ['name', 'urlinfos', 'main_file', 'source'])
# a DatasetInfo Object is basically a tuple of UrlInfos object
# an UrlInfo object is composed of an url and the filenames contained
# in the request content
UrlInfo = namedtuple('UrlInfo', ['url', 'filenames', 'uncompress'])

ROAD_SAFETY_CONFIG = DatasetInfo(
    name='road_safety',
    urlinfos=(
        UrlInfo(
            url="http://data.dft.gov.uk/road-accidents-safety-data/"
                "RoadSafetyData_2015.zip",
            filenames=(
                "Casualties_2015.csv",
                "Vehicles_2015.csv",
                "Accidents_2015.csv"
            ),
            uncompress=True),
        UrlInfo(
            url="http://data.dft.gov.uk/road-accidents-safety-data/"
                "MakeModel2015.zip",
            filenames=("2015_Make_Model.csv",),
            uncompress=True
        )
    ),
    main_file="Accidents_2015.csv",  # for consistency, all files are relevant,
    source="https://data.gov.uk/dataset/road-accidents-safety-data"
)

OPEN_PAYMENTS_CONFIG = DatasetInfo(
    name='open_payments',
    urlinfos=
    (
        UrlInfo(
            url='http://download.cms.gov/openpayments/PGYR13_P011718.ZIP',
            filenames=None, uncompress=True
        ),
    ),
    main_file='OP_DTL_GNRL_PGYR2013_P01172018.csv',  # same
    source='https://openpaymentsdata.cms.gov'
)

MIDWEST_SURVEY_CONFIG = DatasetInfo(
    name='midwest_survey',
    urlinfos=(
        UrlInfo(
            url="https://github.com/fivethirtyeight/data/tree/"
                "master/region-survey/FiveThirtyEight_Midwest_Survey.csv",
            filenames=(
                "FiveThirtyEight_Midwest_Survey.csv",
            ), uncompress=False
        ),
    ),
    main_file="FiveThirtyEight_Midwest_Survey.csv",
    source="https://github.com/fivethirtyeight/data/tree/ master/region-survey"
)
MEDICAL_CHARGE_CONFIG = DatasetInfo(
    name='medical_charge',
    urlinfos=(
        UrlInfo(
            url="https://www.cms.gov/Research-Statistics-Data-and-Systems/"
                "Statistics-Trends-and-Reports/Medicare-Provider-Charge-Data/"
                "Downloads/Inpatient_Data_2011_CSV.zip",
            filenames=(
                "MedicalProviderChargeInpatient.csv",
            ),
            uncompress=True

        ),
    ),
    main_file="MedicalProviderChargeInpatient.csv",
    source="https://www.cms.gov/Research-Statistics-Data-and-Systems/"
           "Statistics-Trends-and-Reports/Medicare-Provider-Charge-Data"
           "/Inpatient.html"
)

EMPLOYEE_SALARIES_CONFIG = DatasetInfo(
    name='employee_salaries',
    urlinfos=(
        UrlInfo(
            url="https://data.montgomerycountymd.gov/api/views/"
                "xj3h-s2i7/rows.csv?accessType=DOWNLOAD",
            filenames=("rows.csv",),
            uncompress=False
        ),
    ),
    main_file="rows.csv",
    source="https://catalog.data.gov/dataset/ employee-salaries-2016"
)

TRAFFIC_VIOLATIONS_CONFIG = DatasetInfo(
    name='traffic_violations',
    urlinfos=(
        UrlInfo(
            url="https://data.montgomerycountymd.gov/api/views/"
                "4mse-ku6q/rows.csv?accessType=DOWNLOAD",
            filenames=(
                "rows.csv",
            ), uncompress=False,
        ),
    ),
    main_file="rows.csv",
    source="https://catalog.data.gov/dataset/ traffic-violations-56dda"
)
FOLDER_PATH = os.path.dirname(os.path.realpath(__file__))


class FileChangedError(Exception):
    pass


def _download_and_write(url, file, show_progress=True):
    if show_progress:  # maybe not an ideal design, should we mark
        # clint as mandatory?
        from clint.textui import progress
    try:
        # using stream=True to download the response body only when
        # accessing the content attribute
        from ..datasets.utils import request_get
        with contextlib.closing(request_get(url, stream=True)) as r:
            total_length = r.headers.get('Content-Length')
            if total_length is not None:
                with open(file, 'wb') as local_file:
                    content_iterator = r.iter_content(chunk_size=1024)
                    if show_progress:
                        content_iterator = progress.bar(
                            content_iterator, expected_size=(int(total_length) /
                                                             1024) + 1)

                    for chunk in content_iterator:
                        if chunk:
                            local_file.write(chunk)
                            local_file.flush()

            else:
                warnings.warn('content size cannot be found, '
                              'downloading file from {} as a whole'.format(
                    url))
                with open(file, 'wb') as local_file:
                    local_file.write(r.content)

    except requests.RequestException as e:
        # pretty general request exception. subject to change
        raise Exception('error while fetching: {}'.format(e))


def fetch_dataset(configfile: DatasetInfo, show_progress=True):
    data_dir = os.path.join(get_data_dir(), configfile.name)
    for urlinfo in configfile.urlinfos:
        _fetch_file(urlinfo.url, data_dir, filenames=urlinfo.filenames,
                    uncompress=urlinfo.uncompress, show_progress=show_progress)
    # returns the absolute path of the csv file where the data is
    result_dict = {
        'description': 'The downloaded data contains the {} dataset.\n'
                       'It can originally be found at: {}'.format(
            configfile.name, configfile.source),
        'path': os.path.join(data_dir, configfile.main_file)
    }
    return result_dict


def _fetch_file(url, data_dir, filenames=None, overwrite=False,
                md5sum=None, uncompress=True, show_progress=True):
    """fetches the content of a requested url

    IF the downloaded file is compressed, then the fetcher
    looks also for the uncompressed files before downloading .


    Parameters
    ----------
    url: str
        url from where to fetch the file from
    data_dir: str
        directory where the data will be stored
    filenames: list
        names of the files in the url content
    overwrite: bool
        whether to overwrite present data
    md5sum: str
        if provided, verifies the integrity of the file using a hash
    uncompress: bool
        whether to uncompress the content of the url

    show_progress: if True, displays a progressbar during the downloading
    of the dataset. Warning: clint needs to be implemented and is not in the
    requirements for now

    Returns
    -------
    a dictionary containing:
    - a short description of the dataset (under the 'description' key )
    - an absolute path leading to the csv file where the data is stored locally
    (under the 'path' key)

    NOTES
    -----
    NON-implemented nilearn parameters:
    * the resume option, that would resume partially downloaded files
    * username/password
    * handlers

    """
    if not os.path.exists(data_dir):
        os.makedirs(data_dir)

    # Determine filename using URL. sticking to urllib.parse, requests does not
    # provide parsing tools
    parse = urllib.parse.urlparse(url)
    file_name = os.path.basename(parse.path)
    download = False
    if file_name == '':
        file_name = md5_hash(parse.path)

    temp_file_name = file_name + ".part"
    full_name = os.path.join(data_dir, file_name)
    temp_full_name = os.path.join(data_dir, temp_file_name)
    if overwrite:
        download = True
        files_to_overwrite = [file_name, temp_file_name]
        if filenames:
            files_to_overwrite += filenames
        for name in files_to_overwrite:
            # remove all compressed/uncompressed files
            _check_if_exists(os.path.join(data_dir, name), remove=True)
    else:
        # if no filenames info provided, overwrite
        is_file_missing = True
        if filenames is not None:
            # first look for uncompressed files
            is_file_missing = any(
                [not _check_if_exists(os.path.join(data_dir, name),
                                      remove=False)
                 for name in filenames])
        if is_file_missing:
            # then look for compressed files
            if not _check_if_exists(full_name, remove=False):
                download = True

    if download:
        _download_and_write(url, temp_full_name, show_progress=show_progress)

    # chunk writing is not implemented, see if necessary
    if _check_if_exists(temp_full_name, remove=False):
        if md5sum is not None:
            if _md5_sum_file(temp_full_name) != md5sum:
                raise FileChangedError(
                    "File %s checksum verification has failed."
                    "Dataset fetching aborted." % temp_full_name)

        shutil.move(temp_full_name, full_name)
    if _check_if_exists(full_name, remove=False) and uncompress:
        _uncompress_file(full_name, delete_archive=True)

    return full_name


def fetch_employee_salaries():
    """fetches the employee_salaries dataset

    The employee_salaries dataset contains information about annual salaries
    (year 2016) for more than 9,000 employees of the Montgomery County
    (Maryland, US).


    Returns
    -------
    dict
        a dictionary containing:
            - a short description of the dataset (under the 'description' key)
            - an absolute path leading to the csv file where the data is
            stored locally (under the 'path' key)


    References
    ----------
    https://catalog.data.gov/dataset/ employee-salaries-2016

    """

    return fetch_dataset(EMPLOYEE_SALARIES_CONFIG, show_progress=False)


def fetch_road_safety():
    """fetches the road safety dataset

    Returns
    -------
    dict
        a dictionary containing:
            - a short description of the dataset (under the 'description' key)
            - an absolute path leading to the csv file where the data is
            stored locally (under the 'path' key)


    References
    ----------
    https://data.gov.uk/dataset/road-accidents-safety-dataset
    """

    return fetch_dataset(ROAD_SAFETY_CONFIG, show_progress=False)


def fetch_medical_charge():
    """fetches the medical charge dataset

    Returns
    -------
    dict
        a dictionary containing:
            - a short description of the dataset (under the 'description' key)
            - an absolute path leading to the csv file where the data is
            stored locally (under the 'path' key)


    References
    ----------
    https://www.cms.gov/Research-Statistics-Data-and-Systems/Statistics-Trends-and-Reports/Medicare-Provider-Charge-Data/Inpatient.html"
    """
    return fetch_dataset(MEDICAL_CHARGE_CONFIG, show_progress=False)


def fetch_midwest_survey():
    """fetches the midwest survey dataset

    Returns
    -------
    dict
        a dictionary containing:
            - a short description of the dataset (under the 'description' key)
            - an absolute path leading to the csv file where the data is
            stored locally (under the 'path' key)


    References
    ----------
    https://github.com/fivethirtyeight/data/tree/master/region-survey
    """
    return fetch_dataset(MIDWEST_SURVEY_CONFIG, show_progress=False)


def fetch_open_payments():
    """fetches the open payements dataset

    Returns
    -------
    dict
        a dictionary containing:
            - a short description of the dataset (under the 'description' key)
            - an absolute path leading to the csv file where the data is
            stored locally (under the 'path' key)


    References
    ----------
    https://openpaymentsdata.cms.gov
    """
    return fetch_dataset(OPEN_PAYMENTS_CONFIG, show_progress=False)


def fetch_traffic_violations():
    """fetches the traffic violations dataset

    Returns
    -------
    dict
        a dictionary containing:
            - a short description of the dataset (under the 'description' key)
            - an absolute path leading to the csv file where the data is
            stored locally (under the 'path' key)


    References
    ----------
    https://catalog.data.gov/dataset/ traffic-violations-56dda
    """
    return fetch_dataset(TRAFFIC_VIOLATIONS_CONFIG, show_progress=False)


if __name__ == '__main__':
    # fetch_midwest_survey()
    fetch_medical_charge()
    # fetch_road_safety()
    # fetch_employee_salaries()
    # fetch_open_payments()



import os
import requests
from io import BytesIO
import zipfile
import contextlib
from functools import wraps


class with_setup:
    def __init__(self, setup, teardown):
        self.setup = setup
        self.teardown = teardown

    def __call__(self, f):
        @wraps(f)
        def func(*args, **kwargs):
            self.setup()
            try:
                f(*args, **kwargs)
            finally:
                self.teardown()

        return func


class MockResponse(requests.Response):
    with_content_length = True
    zipresult = False

    @classmethod
    def set_with_content_length(cls, b):
        """
        set the with_total_length attribute
        Parameters
        ----------
        b: if True, include total_length in the response

        Returns
        -------

        """

        # I added a classmethod for this instead of
        # changing the syntax of request_get by adding
        # a with_total_length in argument that would break the compatibility
        # with request.get
        cls.with_content_length = b

    @classmethod
    def set_to_zipfile(cls, b):
        cls.zipresult = b

    def __init__(self, url, params, content: str, stream=True):
        super(MockResponse, self).__init__()
        self.url = url
        self.params = params
        if self.zipresult:
            # StringIO vs BytesIO: using BytesIO in accorance with the
            # open(file,'wb') from the _fetch_file function
            self.raw = BytesIO()
            with zipfile.ZipFile(self.raw, mode='w',
                                 compression=zipfile.ZIP_DEFLATED) as mf:
                mf.writestr('unzipped_data.txt', 'test')
            # self.raw = self.raw.wr(file=self.raw)
            if stream:
                self._content = self.raw.getvalue()

        else:
            if stream:
                self.raw = BytesIO(str.encode(content))
            else:
                self._content = content

        if self.with_content_length:
            if self.zipresult:
                self.headers['Content-Length'] = len(self.raw.getvalue())
                self.total_length = len(self.raw.getvalue())
            else:
                self.headers['Content-Length'] = len(content)
                self.total_length = len(content)

    def iter_content(self, chunk_size=1, decode_unicode=False):
        """

        it is actually pretty hard to read from a streamed
        zipped file-like object.2 possibilites:
            - force non-stream reading when using zipfiles
            - create a custom iter_content method to handle this case
        Parameters
        ----------
        chunk_size
        decode_unicode

        Returns
        -------

        """

        print('iterating from mocked response')
        return super(MockResponse, self).iter_content(chunk_size=chunk_size,
                                                      decode_unicode=decode_unicode)


def mock_request_get(url, params=None, stream=True):
    # should we be able to input a specialized content for the mockrequest?
    return MockResponse(url, params=None, content=' ')



import collections
import numpy as np

from sklearn.preprocessing import LabelEncoder
from sklearn.base import BaseEstimator, TransformerMixin
from sklearn.utils import check_array


def lambda_(x, n):
    out = x / (x + n)
    return out


class TargetEncoder(BaseEstimator, TransformerMixin):
    """Encode categorical features as a numeric array given a target vector.

    Each category is encoded given the effect that it has in the
    target variable y. The method considers that categorical
    variables can present rare categories. It represents each category by the
    probability of y conditional on this category.
    In addition it takes an empirical Bayes approach to shrink the estimate.

 
    Parameters
    ----------
    categories : 'auto' or a list of lists/arrays of values.
        Categories (unique values) per feature:

        - 'auto' : Determine categories automatically from the training data.
        - list : ``categories[i]`` holds the categories expected in the i-th
          column. The passed categories must be sorted and should not mix
          strings and numeric values.

        The categories used can be found in the ``categories_`` attribute.

    clf_type : string {'regression', 'binary-clf', 'multiclass-clf'}
        The type of classification/regression problem.

    dtype : number type, default np.float64
        Desired dtype of output.

    handle_unknown : 'error' (default) or 'ignore'
        Whether to raise an error or ignore if a unknown categorical feature is
        present during transform (default is to raise). When this parameter
        is set to 'ignore' and an unknown category is encountered during
        transform, the resulting one-hot encoded columns for this feature
        will be all zeros. In the inverse transform, an unknown category
        will be denoted as None.

    Attributes
    ----------
    categories_ : list of arrays
        The categories of each feature determined during fitting
        (in order corresponding with output of ``transform``).

    References
    -----------
    For more details, see Micci-Barreca, 2001: A preprocessing scheme for
    high-cardinality categorical attributes in classification and prediction
    problems.
    """
    def __init__(self,
                 categories='auto',
                 clf_type='binary-clf',
                 dtype=np.float64, handle_unknown='error',
                 ):
        self.categories = categories
        self.dtype = dtype
        self.clf_type = clf_type
        self.handle_unknown = handle_unknown

    def fit(self, X, y):
        """Fit the TargetEncoder to X.
        Parameters
        ----------
        X : array-like, shape [n_samples, n_features]
            The data to determine the categories of each feature.
        y : array
            The associated target vector.

        Returns
        -------
        self
        """
        if self.handle_unknown not in ['error', 'ignore']:
            template = ("handle_unknown should be either 'error' or "
                        "'ignore', got %s")
            raise ValueError(template % self.handle_unknown)

        if self.categories != 'auto':
            for cats in self.categories:
                if not np.all(np.sort(cats) == np.array(cats)):
                    raise ValueError("Unsorted categories are not yet "
                                     "supported")

        X_temp = check_array(X, dtype=None)
        if not hasattr(X, 'dtype') and np.issubdtype(X_temp.dtype, np.str_):
            X = check_array(X, dtype=np.object)
        else:
            X = X_temp

        n_samples, n_features = X.shape

        self._label_encoders_ = [LabelEncoder() for _ in range(n_features)]

        for j in range(n_features):
            le = self._label_encoders_[j]
            Xj = X[:, j]
            if self.categories == 'auto':
                le.fit(Xj)
            else:
                if self.handle_unknown == 'error':
                    valid_mask = np.in1d(Xj, self.categories[j])
                    if not np.all(valid_mask):
                        diff = np.unique(Xj[~valid_mask])
                        msg = ("Found unknown categories {0} in column {1}"
                               " during fit".format(diff, j))
                        raise ValueError(msg)
                le.classes_ = np.array(self.categories[j])

        self.categories_ = [le.classes_ for le in self._label_encoders_]
        self.n = len(y)
        if self.clf_type in ['binary-clf', 'regression']:
            self.Eyx_ = [{cat: np.mean(y[X[:, j] == cat])
                          for cat in self.categories_[j]}
                         for j in range(len(self.categories_))]
            self.Ey_ = np.mean(y)
            self.counter_ = {j: collections.Counter(X[:, j])
                             for j in range(n_features)}
        if self.clf_type in ['multiclass-clf']:
            self.classes_ = np.unique(y)

            self.Eyx_ = {c: [{cat: np.mean((y == c)[X[:, j] == cat])
                              for cat in self.categories_[j]}
                             for j in range(len(self.categories_))]
                         for c in self.classes_}
            self.Ey_ = {c: np.mean(y == c) for c in self.classes_}
            self.counter_ = {j: collections.Counter(X[:, j])
                             for j in range(n_features)}
        self.k = {j: len(self.counter_[j]) for j in self.counter_}
        return self

    def transform(self, X):
        """Transform X using specified encoding scheme.

        Parameters
        ----------
        X : array-like, shape [n_samples, n_features_new]
            The data to encode. 

        Returns
        -------
        X_new : 2-d array
            Transformed input.
        """
        X_temp = check_array(X, dtype=None)
        if not hasattr(X, 'dtype') and np.issubdtype(X_temp.dtype, np.str_):
            X = check_array(X, dtype=np.object)
        else:
            X = X_temp

        n_samples, n_features = X.shape
        X_int = np.zeros_like(X, dtype=np.int)
        X_mask = np.ones_like(X, dtype=np.bool)

        for i in range(n_features):
            Xi = X[:, i]
            valid_mask = np.in1d(Xi, self.categories_[i])

            if not np.all(valid_mask):
                if self.handle_unknown == 'error':
                    diff = np.unique(X[~valid_mask, i])
                    msg = ("Found unknown categories {0} in column {1}"
                           " during transform".format(diff, i))
                    raise ValueError(msg)
                else:
                    # Set the problematic rows to an acceptable value and
                    # continue `The rows are marked `X_mask` and will be
                    # removed later.
                    X_mask[:, i] = valid_mask
                    Xi = Xi.copy()
                    Xi[~valid_mask] = self.categories_[i][0]
            X_int[:, i] = self._label_encoders_[i].transform(Xi)

        out = []
        for j, cats in enumerate(self.categories_):
            unqX = np.unique(X[:, j])
            encoder = {x: 0 for x in unqX}
            if self.clf_type in ['binary-clf', 'regression']:
                for x in unqX:
                    if x not in cats:
                        Eyx = 0
                    else:
                        Eyx = self.Eyx_[j][x]
                    lambda_n = lambda_(self.counter_[j][x], self.n/self.k[j])
                    encoder[x] = lambda_n*Eyx + (1 - lambda_n)*self.Ey_
                x_out = np.zeros((len(X[:, j]), 1))
                for i, x in enumerate(X[:, j]):
                    x_out[i, 0] = encoder[x]
                out.append(x_out.reshape(-1, 1))
            if self.clf_type == 'multiclass-clf':
                x_out = np.zeros((len(X[:, j]), len(self.classes_)))
                lambda_n = {x: 0 for x in unqX}
                for x in unqX:
                    lambda_n[x] = lambda_(self.counter_[j][x], self.n/self.k[j])
                for k, c in enumerate(np.unique(self.classes_)):
                    for x in unqX:
                        if x not in cats:
                            Eyx = 0
                        else:
                            Eyx = self.Eyx_[c][j][x]
                        encoder[x] = lambda_n[x]*Eyx + \
                            (1 - lambda_n[x])*self.Ey_[c]
                    for i, x in enumerate(X[:, j]):
                        x_out[i, k] = encoder[x]
                out.append(x_out)
        out = np.hstack(out)
        return out



"""
Some string distances
"""
import functools

import numpy as np

try:
    import Levenshtein
    _LEVENSHTEIN_AVAILABLE = True
except ImportError:
    _LEVENSHTEIN_AVAILABLE = False

from collections import Counter
# Levenstein, adapted from
# https://en.wikibooks.org/wiki/Algorithm_Implementation/Strings/Levenshtein_distance#Python


# TODO vectorize these functions (accept arrays)


def levenshtein_array(source, target):
    target_size = len(target)
    if len(source) < target_size:
        return levenshtein_array(target, source)

    # So now we have len(source) >= len(target).
    if len(target) == 0:
        return len(source)

    # Create numpy arrays
    source = np.array(tuple(source), dtype='|U1')
    target = np.array(tuple(target), dtype='|U1')

    # We use a dynamic programming algorithm, but with the
    # added optimization that we only need the last two rows
    # of the matrix.
    previous_row = np.arange(target_size + 1)
    for s in source:
        # Insertion (target grows longer than source):
        current_row = previous_row + 1

        # Substitution or matching:
        # Target and source items are aligned, and either
        # are different (cost of 1), or are the same (cost of 0).
        current_row[1:] = np.minimum(
                current_row[1:],
                np.add(previous_row[:-1], target != s))

        # Deletion (target grows shorter than source):
        current_row[1:] = np.minimum(
                current_row[1:],
                current_row[0:-1] + 1)

        previous_row = current_row

    return previous_row[-1]


def levenshtein_seq(seq1, seq2):
    thisrow = list(range(1, len(seq2) + 1)) + [0]
    len_seq2 = len(seq2)
    for x in range(len(seq1)):
        oneago = thisrow
        thisrow = [0] * len_seq2 + [x + 1]
        for y in range(len_seq2):
            delcost = oneago[y] + 1
            addcost = thisrow[y - 1] + 1
            subcost = oneago[y - 1] + (seq1[x] != seq2[y])
            thisrow[y] = min(delcost, addcost, subcost)
    return thisrow[len_seq2 - 1]


def levenshtein(seq1, seq2):
    # Choose the fastest option depending on the size of the arrays
    # The number 15 was chosen empirically on Python 3.6
    if _LEVENSHTEIN_AVAILABLE:
        return Levenshtein.distance(seq1, seq2)
    if len(seq1) < 15:
        return levenshtein_seq(seq1, seq2)
    else:
        return levenshtein_array(seq1, seq2)


def _levenshtein_ratio(seq1, seq2):
    # Private function not using the Levenshtein package
    total_len = len(seq1) + len(seq2)
    if total_len == 0:
        return 1.
    return (total_len - levenshtein(seq1, seq2)) / total_len


if _LEVENSHTEIN_AVAILABLE:
    levenshtein_ratio = Levenshtein.ratio
else:
    levenshtein_ratio = _levenshtein_ratio


def _jaro_winkler(seq1, seq2, winkler=False):
    # Adapted from the jellyfish package
    # If winkler is False, the Jaro similarity is returned, else the
    # Jaro-Winkler variant is returned

    seq1_len = len(seq1)
    seq2_len = len(seq2)

    if not seq1_len or not seq2_len:
        return 0.0

    min_len = max(seq1_len, seq2_len)
    search_range = (min_len // 2) - 1
    if search_range < 0:
        search_range = 0

    seq1_flags = seq1_len * [False]
    seq2_flags = seq2_len * [False]

    # looking only within search range, count & flag matched pairs
    common_chars = 0
    for i, seq1_ch in enumerate(seq1):
        low = i - search_range if i > search_range else 0
        hi = i + search_range if i + search_range < seq2_len else seq2_len - 1
        for j in range(low, hi + 1):
            if not seq2_flags[j] and seq2[j] == seq1_ch:
                seq1_flags[i] = seq2_flags[j] = True
                common_chars += 1
                break

    # short circuit if no characters match
    if not common_chars:
        return 0.0

    # count transpositions
    k = trans_count = 0
    for i, seq1_f in enumerate(seq1_flags):
        if seq1_f:
            for j in range(k, seq2_len):
                if seq2_flags[j]:
                    k = j + 1
                    break
            if seq1[i] != seq2[j]:
                trans_count += 1
    trans_count /= 2

    # adjust for similarities in nonmatched characters
    common_chars = float(common_chars)
    weight = ((common_chars/seq1_len + common_chars/seq2_len +
              (common_chars-trans_count) / common_chars)) / 3

    # winkler modification: continue to boost if strings are similar
    if winkler and weight > 0.7 and seq1_len > 3 and seq2_len > 3:
        # adjust for up to first 4 chars in common
        j = min(min_len, 4)
        i = 0
        while i < j and seq1[i] == seq2[i] and seq1[i]:
            i += 1
        if i:
            weight += i * 0.1 * (1.0 - weight)

    return weight


if _LEVENSHTEIN_AVAILABLE:
    jaro_winkler = Levenshtein.jaro_winkler
    jaro = Levenshtein.jaro
else:
    jaro_winkler = functools.partial(_jaro_winkler, winkler=True)
    jaro = _jaro_winkler


def get_unique_ngrams(string, n):
    """ Return the set of different tri-grams in a string
    """
    spaces = ' '  # * (n // 2 + n % 2)
    string = spaces + " ".join(string.lower().split()) + spaces
    string_list = [string[i:] for i in range(n)]
    return set(zip(*string_list))


def get_ngrams(string, n):
    """ Return the set of different tri-grams in a string
    """
    # Pure Python implementation: no numpy
    spaces = ' '  # * (n // 2 + n % 2)
    string = spaces + " ".join(string.lower().split()) + spaces
    string_list = [string[i:] for i in range(n)]
    return list(zip(*string_list))


def ngram_similarity(string1, string2, n):
    """ n-gram similarity between two strings
    """
    ngrams1 = get_ngrams(string1, n)
    count1 = Counter(ngrams1)

    ngrams2 = get_ngrams(string2, n)
    count2 = Counter(ngrams2)

    samegrams = sum((count1 & count2).values())
    allgrams = len(ngrams1) + len(ngrams2)
    similarity = samegrams/(allgrams - samegrams)
    return similarity


if __name__ == '__main__':
    s1 = 'aa'
    s2 = 'aaab'
    print('Levenshtein similarity: %.3f' % levenshtein(s1, s2))
    print('3-gram similarity: %.3f' % ngram_similarity(s1, s2, 3))



import numpy as np

from dirty_cat import similarity_encoder, string_distances
from dirty_cat.similarity_encoder import get_kmeans_prototypes


def _test_similarity(similarity, similarity_f, hashing_dim=None, categories='auto', n_prototypes=None):
    if n_prototypes is None:
        X = np.array(['aa', 'aaa', 'aaab']).reshape(-1, 1)
        X_test = np.array([['Aa', 'aAa', 'aaa', 'aaab', ' aaa  c']]).reshape(-1, 1)

        model = similarity_encoder.SimilarityEncoder(
            similarity=similarity, handle_unknown='ignore',
            hashing_dim=hashing_dim, categories=categories,
            n_prototypes=n_prototypes)

        if similarity == 'ngram':
            model.ngram_range = (3, 3)

        encoder = model.fit(X).transform(X_test)

        ans = np.zeros((len(X_test), len(X)))
        for i, x_t in enumerate(X_test.reshape(-1)):
            for j, x in enumerate(X.reshape(-1)):
                if similarity == 'ngram':
                    ans[i, j] = similarity_f(x_t, x, 3)
                else:
                    ans[i, j] = similarity_f(x_t, x)
        assert np.array_equal(encoder, ans)
    else:
        X = np.array(
            ['aac', 'aaa', 'aaab', 'aaa', 'aaab', 'aaa', 'aaab', 'aaa']
        ).reshape(-1, 1)
        X_test = np.array([['Aa', 'aAa', 'aaa', 'aaab', ' aaa  c']]
                          ).reshape(-1, 1)

        try:
            model = similarity_encoder.SimilarityEncoder(
                similarity=similarity, handle_unknown='ignore',
                hashing_dim=hashing_dim, categories=categories,
                n_prototypes=n_prototypes, random_state=42)
        except ValueError as e:
            assert (e.__str__() == 'n_prototypes expected None or a positive non null integer')
            return

        if similarity == 'ngram':
            model.ngram_range = (3, 3)

        encoder = model.fit(X).transform(X_test)
        if n_prototypes == 1:
            assert (model.categories_ == ['aaa'])
        elif n_prototypes == 2:
            a = [np.array(['aaa', 'aaab'], dtype='<U4')]
            assert (np.array_equal(a, model.categories_))
        elif n_prototypes == 3:
            a = [np.array(['aaa', 'aaab', 'aac'], dtype='<U4')]
            assert (np.array_equal(a, model.categories_))

        ans = np.zeros((len(X_test),
                        len(np.array(model.categories_).reshape(-1))))
        for i, x_t in enumerate(X_test.reshape(-1)):
            for j, x in enumerate(np.array(model.categories_).reshape(-1)):
                if similarity == 'ngram':
                    ans[i, j] = similarity_f(x_t, x, 3)
                else:
                    ans[i, j] = similarity_f(x_t, x)

        assert np.array_equal(encoder, ans)


def test_similarity_encoder():
    categories = ['auto', 'most_frequent', 'k-means']
    for category in categories:
        if category == 'auto':
            _test_similarity('levenshtein-ratio',
                             string_distances.levenshtein_ratio,
                             categories=category,
                             n_prototypes=None)
            _test_similarity('jaro-winkler', string_distances.jaro_winkler,
                             categories=category, n_prototypes=None)
            _test_similarity('jaro', string_distances.jaro,
                             categories=category, n_prototypes=None)
            _test_similarity('ngram', string_distances.ngram_similarity,
                             categories=category, n_prototypes=None)
            _test_similarity('ngram', string_distances.ngram_similarity,
                             hashing_dim=2 ** 16, categories=category)
        else:
            for i in range(1, 4):
                _test_similarity('levenshtein-ratio',
                                 string_distances.levenshtein_ratio,
                                 categories=category,
                                 n_prototypes=i)
                _test_similarity('jaro-winkler', string_distances.jaro_winkler,
                                 categories=category, n_prototypes=i)
                _test_similarity('jaro', string_distances.jaro,
                                 categories=category, n_prototypes=i)
                _test_similarity('ngram', string_distances.ngram_similarity,
                                 categories=category, n_prototypes=i)
                _test_similarity('ngram', string_distances.ngram_similarity,
                                 hashing_dim=2 ** 16, categories=category,
                                 n_prototypes=i)


def test_kmeans_protoypes():
    X_test = np.array(['cbbba', 'baaac', 'accc'])
    proto = get_kmeans_prototypes(X_test, 3)
    assert np.array_equal(np.sort(proto), np.sort(X_test))


def test_reproducibility():
    sim_enc = similarity_encoder.SimilarityEncoder(categories='k-means', n_prototypes=10, random_state=435)
    X = np.array([' %s ' % chr(i) for i in range(32, 127)]).reshape((-1, 1))
    prototypes = sim_enc.fit(X).categories_[0]
    for i in range(10):
        assert (np.array_equal(prototypes, sim_enc.fit(X).categories_[0]))



"""
dirty_cat: Learning on dirty categories.
"""
import os

version_file = os.path.join(os.path.dirname(__file__), 'VERSION.txt')
with open(version_file) as fh:
    __version__ = fh.read().strip()

from .similarity_encoder import SimilarityEncoder
from .target_encoder import TargetEncoder

__all__ = ['SimilarityEncoder', 'TargetEncoder']



"""
datasets module of dirty cat
"""

from .fetching import fetch_midwest_survey, fetch_employee_salaries, \
    fetch_open_payments, fetch_road_safety, fetch_medical_charge, \
    fetch_traffic_violations
from .fetching import get_data_dir

__all__ = [
    'get_data_dir',
    'fetch_medical_charge',
    'fetch_midwest_survey',
    'fetch_employee_salaries',
    'fetch_road_safety',
    'fetch_open_payments',
    'fetch_traffic_violations'
]



"""
Test the datasets module
"""
# Author: Pierre Glaser
# -*- coding: utf-8 -*-
import os
from tempfile import mkstemp
import shutil
import zipfile
import contextlib

import dirty_cat.datasets.utils as datasets_utils
import dirty_cat.datasets.tests.utils as utils

currdir = os.path.dirname(os.path.abspath(__file__))
datadir = os.path.join(currdir, 'data')
tmpdir = None
url_request = None
file_mock = None


def setup_mock(true_module=datasets_utils, mock_module=utils):
    global original_url_request
    global mock_url_request
    mock_url_request = mock_module.mock_request_get
    original_url_request = true_module.request_get
    true_module.request_get = mock_url_request


def teardown_mock(true_module=datasets_utils):
    global original_url_request
    true_module.request_get = original_url_request


@utils.with_setup(setup=setup_mock, teardown=teardown_mock)
def test_fetch_file_overwrite():
    utils.MockResponse.set_with_content_length(True)
    utils.MockResponse.set_to_zipfile(False)
    test_dir = datasets_utils.get_data_dir(name='test')
    from dirty_cat.datasets import fetching
    try:
        # test that filename is a md5 hash of the url if
        # the url ends with /
        fil = fetching._fetch_file(url='http://foo/', data_dir=test_dir,
                                   overwrite=True, uncompress=False,
                                   show_progress=False)
        assert os.path.basename(fil) == datasets_utils.md5_hash('/')
        os.remove(fil)

        # overwrite non-exiting file.
        fil = fetching._fetch_file(url='http://foo/testdata', data_dir=test_dir,
                                   overwrite=True, uncompress=False,
                                   show_progress=False)

        # check if data_dir is actually used
        assert os.path.dirname(fil) == test_dir
        assert os.path.exists(fil)
        with open(fil, 'r') as fp:
            assert fp.read() == ' '

        # Modify content
        with open(fil, 'w') as fp:
            fp.write('some content')

        # Don't overwrite existing file.
        fil = fetching._fetch_file(url='http://foo/testdata', data_dir=test_dir,
                                   overwrite=False, uncompress=False,
                                   show_progress=False)
        assert os.path.exists(fil)
        with open(fil, 'r') as fp:
            assert fp.read() == 'some content'

        # Overwrite existing file.
        # Overwrite existing file.
        fil = fetching._fetch_file(url='http://foo/testdata', data_dir=test_dir,
                                   overwrite=True, uncompress=False,
                                   show_progress=False)
        assert os.path.exists(fil)
        with open(fil, 'r') as fp:
            assert fp.read() == ' '

        # modify content,
        # change filename,add it in argument, and set overwrite to false
        with open(fil, 'w') as fp:
            fp.write('some content')
        newf = 'moved_file'
        os.rename(fil, os.path.join(test_dir, newf))
        fetching._fetch_file(url='http://foo/testdata',
                             filenames=(newf,),
                             data_dir=test_dir,
                             overwrite=False, uncompress=False,
                             show_progress=False)
        assert (
            not os.path.exists(fil))  # it has been removed and should not have
        with open(os.path.join(test_dir, newf), 'r') as fp:
            assert fp.read() == 'some content'
        # been downloaded again
        os.remove(os.path.join(test_dir, newf))

        # # create a zipfile with a file inside, remove the file, and
        #
        # zipd = os.path.join('testzip.zip')
        # with contextlib.closing(zipfile.ZipFile())
        #     fetching._fetch_file(url='http://foo/', filenames=('test_filename',),
        #                          data_dir=test_dir,
        #                          overwrite=False, uncompress=False,
        #                          show_progress=False)

        # add wrong md5 sum file and catch ValueError
        try:
            fil = fetching._fetch_file(url='http://foo/testdata',
                                       data_dir=test_dir,
                                       overwrite=True, uncompress=False,
                                       show_progress=False, md5sum='1')
            raise ValueError  # if no error raised in the previous line,
            #  it is bad:
            # a wrong md5 sum should raise an error. So forcing the except chunk
            # to happen anyway
        except Exception as e:
            assert isinstance(e, fetching.FileChangedError)
        utils.MockResponse.set_with_content_length(False)
        # write content if no content size
        fil = fetching._fetch_file(url='http://foo/testdata', data_dir=test_dir,
                                   overwrite=True, uncompress=False,
                                   show_progress=False)
        assert os.path.exists(fil)
        os.remove(fil)
    finally:
        shutil.rmtree(test_dir)


def test_md5_sum_file():
    # Create dummy temporary file
    out, f = mkstemp()
    os.write(out, b'abcfeg')
    os.close(out)
    assert datasets_utils._md5_sum_file(f) == '18f32295c556b2a1a3a8e68fe1ad40f7'
    os.remove(f)


@utils.with_setup(setup=setup_mock, teardown=teardown_mock)
def test_fetch_dataset():
    from dirty_cat.datasets import fetching
    utils.MockResponse.set_with_content_length(True)
    datadir = os.path.join(datasets_utils.get_data_dir(),
                           'testdata')
    try:
        urlinfo = fetching.DatasetInfo(name='testdata',
                                       urlinfos=(fetching.UrlInfo(
                                           url='http://foo/data',
                                           filenames=('data',),
                                           uncompress=False),),
                                       main_file='data',
                                       source='http://foo/')
        fetching.fetch_dataset(urlinfo, show_progress=False)
        assert os.path.exists(os.path.join(datadir, 'data'))
        shutil.rmtree(os.path.join(datadir))

        # test with zipped data
        utils.MockResponse.set_to_zipfile(True)
        utils.MockResponse.set_with_content_length(False)
        urlinfo = fetching.DatasetInfo(name='testdata',
                                       urlinfos=(fetching.UrlInfo(
                                           url='http://foo/data.zip',
                                           filenames=('unzipped_data.txt',),
                                           uncompress=True),),
                                       main_file='data',
                                       source='http://foo/')
        fetching.fetch_dataset(urlinfo, show_progress=False)
        assert os.path.exists(os.path.join(datadir, 'unzipped_data.txt'))
    finally:
        if os.path.exists(datadir):
            shutil.rmtree(datadir)


if __name__ == '__main__':
    # pass
    # test_fetch_file_overwrite()
    test_fetch_dataset()



"""
Benchmark time consumption and scores for K-means and most frequent strategies.

We use the traffic_violations dataset to benchmark the different dimensionality
reduction strategies used in similarity encoding.

Parameters that are modified:
- Number of rows in datasets: 10k, 20k, 50k, 100k and nuniques.
- Hashing dimensions: 2 ** 14, 2 ** 16, 2 ** 18, 2 ** 20
- Ngram-range: (3, 3), (2, 4)
"""

# We filter out the warning asking us to specify the solver in the logistic_regression
import warnings
warnings.simplefilter(action='ignore', category=FutureWarning)

from time import time

import matplotlib.pyplot as plt
import pandas as pd
import seaborn as sns
from sklearn import pipeline, linear_model, model_selection
from sklearn.compose import ColumnTransformer
from sklearn.preprocessing import OneHotEncoder

from dirty_cat import SimilarityEncoder
from dirty_cat.datasets import fetch_traffic_violations


data = fetch_traffic_violations()
dfr = pd.read_csv(data['path'])

transformers = [
    ('one_hot', OneHotEncoder(sparse=False, handle_unknown='ignore'),
     ['Alcohol',
      'Arrest Type',
      'Belts',
      'Commercial License',
      'Commercial Vehicle',
      'Fatal',
      'Gender',
      'HAZMAT',
      'Property Damage',
      'Race',
      'Work Zone']),
    ('pass', 'passthrough', ['Year']),
]


def benchmark(strat='k-means', limit=50000, n_proto=100, hash_dim=None, ngram_range=(3, 3)):
    df = dfr[:limit].copy()
    df = df.dropna(axis=0)
    df = df.reset_index()

    y = df['Violation Type']

    if strat == 'k-means':
        sim_enc = SimilarityEncoder(similarity='ngram', ngram_range=ngram_range, categories='k-means',
                                    hashing_dim=hash_dim, n_prototypes=n_proto, random_state=3498)
    else:
        sim_enc = SimilarityEncoder(similarity='ngram', ngram_range=ngram_range, categories='most_frequent',
                                    hashing_dim=hash_dim, n_prototypes=n_proto, random_state=3498)

    column_trans = ColumnTransformer(
        transformers=transformers + [('sim_enc', sim_enc, ['Description'])],
        remainder='drop'
    )

    t0 = time()
    X = column_trans.fit_transform(df)
    t1 = time()
    t_score_1 = t1 - t0

    model = pipeline.Pipeline([('logistic', linear_model.LogisticRegression())])

    t0 = time()
    m_score = model_selection.cross_val_score(model, X, y, cv=20)
    t1 = time()
    t_score_2 = t1 - t0
    return t_score_1, m_score, t_score_2


def plot(bench, title=''):
    sns.set(style='ticks', palette='muted')
    hash_dims = ['Count', '2 ** 14', '2 ** 16', '2 ** 18', '2 ** 20']
    scores = []
    vectorizer = []
    strategy = []

    for i, e in enumerate(bench):
        vectorizer.extend([hash_dims[i % 5]] * (2 * len(e[0][1])))
        strategy.extend(['k-means'] * len(e[0][1]))
        strategy.extend(['most-frequent'] * len(e[1][1]))
        scores.extend(e[0][1])
        scores.extend(e[1][1])

    df = pd.DataFrame(columns=['vectorizer', 'strategy', 'score'])
    df['vectorizer'] = vectorizer
    df['strategy'] = strategy
    df['score'] = scores

    first = plt.figure()
    ax = sns.boxplot(x='vectorizer', y='score', hue='strategy', data=df)
    ax.set(title=title, xlabel='Vectorizer used', ylabel='Mean score on 10 cross validations')
    ax.legend(loc='center left', bbox_to_anchor=(1, 0.5))
    first.tight_layout()

    vectorizer.clear()
    scores.clear()
    strategy.clear()
    times = []

    for i, e in enumerate(bench):
        vectorizer.extend([hash_dims[i % 5]] * 4)
        strategy.extend(['K-means vect', 'K-means X-val', 'MF vect', 'MF X-val'])
        times.extend([e[0][0], e[0][2] / 20, e[1][0], e[1][2] / 20])

    df = pd.DataFrame(columns=['vectorizer', 'strategy/operation', 'time'])
    df['vectorizer'] = vectorizer
    df['strategy/operation'] = strategy
    df['time'] = times

    second = plt.figure()
    ax1 = sns.barplot(x='vectorizer', y='time', hue='strategy/operation', data=df)
    ax1.set(title=title, xlabel='Vectorizer used', ylabel='Time in seconds')
    ax1.legend(loc='center left', bbox_to_anchor=(1, 0.5))
    second.tight_layout()

    title = title.replace(' ', '_').replace(':', '-').replace(',', '_').lower()
    first.savefig(title + '_score.png')
    second.savefig(title + '_time.png')
    # first.show()
    # second.show(t)


def loop(proto):
    limits = sorted([dfr['Description'].nunique(), 10000, 20000, 50000, 100000])
    hash_dims = [None, 2 ** 14, 2 ** 16, 2 ** 18, 2 ** 20]
    bench = list()
    ngram = [(3, 3), (2, 4)]
    for limit in limits:
        for r in ngram:
            for h in hash_dims:
                bench.append((benchmark(strat='k-means', limit=limit, n_proto=proto, hash_dim=h, ngram_range=r),
                              benchmark(strat='most-frequent', limit=limit, n_proto=proto, hash_dim=h, ngram_range=r)))
            title = 'N-gram range: %s, Rows: %d, Prototypes: %d, 20 CV' % (r.__str__(), limit, proto)
            plot(bench, title)
            bench.clear()


if __name__ == '__main__':
    loop(100)



import numpy as np

from dirty_cat import target_encoder


def test_similarity_encoder():
    lambda_ = target_encoder.lambda_
    X1 = np.array(['Red',
                   'red',
                   'green',
                   'blue',
                   'green',
                   'green',
                   'blue',
                   'red']).reshape(-1, 1)
    X2 = np.array(['male',
                   'male',
                   'female',
                   'male',
                   'female',
                   'female',
                   'female',
                   'male']).reshape(-1, 1)
    X = np.hstack([X1, X2])

    # Case 1: binary-classification and regression
    y = np.array([1, 0, 0, 1, 0, 1, 0, 0])
    n = len(y)

    Ey_ = 3/8
    Eyx_ = {'color': {'Red': 1,
                      'red': 0,
                      'green': 1/3,
                      'blue': .5},
            'gender': {'male': .5,
                       'female': .25}}
    count_ = {'color': {'Red': 1,
                        'red': 2,
                        'green': 3,
                        'blue': 2},
              'gender': {'male': 4,
                         'female': 4}}

    encoder = target_encoder.TargetEncoder()
    encoder.fit(X, y)
    for j in range(X.shape[1]):
        assert np.array_equal(encoder.categories_[j], np.unique(X[:, j]))
    assert Ey_ == encoder.Ey_
    assert encoder.Eyx_[0] == Eyx_['color']
    assert encoder.Eyx_[1] == Eyx_['gender']

    Xtest1 = np.array(['Red',
                       'red',
                       'blue',
                       'green',
                       'Red',
                       'red',
                       'blue',
                       'green']).reshape(-1, 1)
    Xtest2 = np.array(['male',
                       'male',
                       'male',
                       'male',
                       'female',
                       'female',
                       'female',
                       'female']).reshape(-1, 1)
    Xtest = np.hstack([Xtest1, Xtest2])

    Xout = encoder.transform(Xtest)

    ans_dict = {var:
                {cat:
                 Eyx_[var][cat] *
                 lambda_(count_[var][cat], n/len(count_[var])) +
                 Ey_ * (1 - lambda_(count_[var][cat], n/len(count_[var])))
                 for cat in Eyx_[var]}
                for var in Eyx_
                }
    ans = np.zeros((n, 2))
    for j, col in enumerate(['color', 'gender']):
        for i in range(n):
            ans[i, j] = ans_dict[col][Xtest[i, j]]
    assert np.array_equal(Xout, ans)

    # Case 2: multiclass-classification
    y = np.array([1, 0, 2, 1, 0, 1, 0, 0])
    n = len(y)

    encoder = target_encoder.TargetEncoder(clf_type='multiclass-clf')
    encoder.fit(X, y)

    Ey_ = {0: 4/8, 1: 3/8, 2: 1/8}
    Eyx_ = {0: {}, 1: {}, 2: {}}
    Eyx_[0] = {'color': {'Red': 0,
                         'red': 1,
                         'green': 1/3,
                         'blue': .5},
               'gender': {'male': .5,
                          'female': .5}}
    Eyx_[1] = {'color': {'Red': 1,
                         'red': 0,
                         'green': 1/3,
                         'blue': .5},
               'gender': {'male': .5,
                          'female': .25}}
    Eyx_[2] = {'color': {'Red': 0,
                         'red': 0,
                         'green': 1/3,
                         'blue': 0},
               'gender': {'male': 0,
                          'female': .25}}
    assert np.array_equal(np.unique(y), encoder.classes_)
    for k in [0, 1, 2]:
        assert Ey_[k] == encoder.Ey_[k]
        assert encoder.Eyx_[k][0] == Eyx_[k]['color']
        assert encoder.Eyx_[k][1] == Eyx_[k]['gender']

    count_ = {'color': {'Red': 1,
                        'red': 2,
                        'green': 3,
                        'blue': 2},
              'gender': {'male': 4,
                         'female': 4}}
    assert count_['color'] == encoder.counter_[0]
    assert count_['gender'] == encoder.counter_[1]

    ans_dict = {0: {}, 1: {}, 2: {}}
    for k in [0, 1, 2]:
        ans_dict[k] = {var:
                       {cat:
                        Eyx_[k][var][cat] *
                        lambda_(count_[var][cat], n/len(count_[var])) +
                        Ey_[k] *
                        (1 - lambda_(count_[var][cat], n/len(count_[var])))
                        for cat in Eyx_[k][var]}
                       for var in Eyx_[k]
                       }

    ans = np.zeros((n, 2*3))
    Xtest1 = np.array(['Red',
                       'red',
                       'blue',
                       'green',
                       'Red',
                       'red',
                       'blue',
                       'green']).reshape(-1, 1)
    Xtest2 = np.array(['male',
                       'male',
                       'male',
                       'male',
                       'female',
                       'female',
                       'female',
                       'female']).reshape(-1, 1)
    Xtest = np.hstack([Xtest1, Xtest2])
    for k in [0, 1, 2]:
        for j, col in enumerate(['color', 'gender']):
            for i in range(n):
                ans[i, 2*j+k] = ans_dict[k][col][Xtest[i, j]]

    Xout = encoder.transform(Xtest)
    ans = np.zeros((n, 2*3))
    Xtest1 = np.array(['Red',
                       'red',
                       'blue',
                       'green',
                       'Red',
                       'red',
                       'blue',
                       'green']).reshape(-1, 1)
    Xtest2 = np.array(['male',
                       'male',
                       'male',
                       'male',
                       'female',
                       'female',
                       'female',
                       'female']).reshape(-1, 1)
    Xtest = np.hstack([Xtest1, Xtest2])
    for k in [0, 1, 2]:
        for j, col in enumerate(['color', 'gender']):
            for i in range(n):
                ans[i, j*3+k] = ans_dict[k][col][Xtest[i, j]]
    encoder = target_encoder.TargetEncoder(clf_type='multiclass-clf')
    encoder.fit(X, y)
    Xout = encoder.transform(Xtest)
    assert np.array_equal(Xout, ans)



"""
Fast counting of 3-grams for short strings.


Quick benchmarking seems to show that pure Python code is faster when
for strings less that 1000 characters, and numpy versions is faster for
longer strings.

Very long strings would benefit from probabilistic counting (bloom
filter, count min sketch) as implemented eg in the "bounter" module.

Run unit tests with "pytest count_3_grams.py"
"""

###############################################################################


def get_unique_3grams(string):
    """ Return the set of different tri-grams in a string
    """
    # Pure Python implementation: no numpy
    string = '  ' + string + '  '
    return set(zip(string, string[1:], string[2:]))


def get_3grams(string):
    """ Return the set of different tri-grams in a string
    """
    # Pure Python implementation: no numpy
    string = '  ' + string + '  '
    return list(zip(string, string[1:], string[2:]))


def number_of_common_3grams(string1, string2):
    """ Return the number of common tri-grams in two strings
    """
    # Pure Python implementation: no numpy
    tri_grams = set(zip(string1, string1[1:], string1[2:]))
    tri_grams = tri_grams.intersection(zip(string2, string2[1:],
                                           string2[2:]))
    return len(tri_grams)


def test_number_of_3grams():
    # Small unit tests
    assert number_of_3grams('abc') == 1
    for i in range(1, 7):
        assert number_of_3grams(i * 'aaa') == 1
        assert number_of_3grams('abcdefghifk'[:i+2]) == i


def test_number_of_common_3grams():
    # Small unit tests
    for i in range(1, 7):
        assert number_of_common_3grams(i * 'aaa', i * 'aaa') == 1
        assert number_of_common_3grams('abcdefghifk'[:i+2],
                                       'abcdefghifk'[:i+2]) == i


###############################################################################
# Numpy versions
import numpy as np


def number_of_3grams_np(string):
    """ Return the number of different tri-grams in a string
    """
    arr = np.frombuffer(string, dtype='S1')

    # Hard coding the 3 gram
    # Appending 3 shifted versions of the strings
    arr = np.concatenate((arr[:-2, None], arr[1:-1, None], arr[2:, None]),
                         axis=1)
    # "concatenate" strings
    arr = arr.view('S3')
    return np.unique(arr).size


def number_of_common_3grams_np(string1, string2):
    """ Return the number of common tri-grams in two strings
    """
    arr1 = np.frombuffer(string1, dtype='S1')
    arr2 = np.frombuffer(string2, dtype='S1')

    # Hard coding the 3 gram
    # Appending 3 shifted versions of the strings
    arr1 = np.concatenate((arr1[:-2, None], arr1[1:-1, None], arr1[2:, None]),
                          axis=1)
    # "concatenate" strings
    arr1 = arr1.view('S3')

    arr2 = np.concatenate((arr2[:-2, None], arr2[1:-1, None], arr2[2:, None]),
                          axis=1)
    arr2 = arr2.view('S3')

    return np.lib.arraysetops.intersect1d(arr1, arr2).size


def test_number_of_3grams_np():
    # Small unit tests
    assert number_of_3grams_np('abc') == 1
    for i in range(1, 7):
        assert number_of_3grams_np(i * 'aaa') == 1
        assert number_of_3grams_np('abcdefghifk'[:i+2]) == i


def test_number_of_common_3grams_np():
    # Small unit tests
    for i in range(1, 7):
        assert number_of_common_3grams_np(i * 'aaa', i * 'aaa') == 1
        assert number_of_common_3grams_np('abcdefghifk'[:i+2],
                                          'abcdefghifk'[:i+2]) == i

###############################################################################


if __name__ == '__main__':
    # Our two example strings
    s1 = 'patricio'
    s2 = 'patricia'

    print(number_of_common_3grams(s1, s2))



import sys
import os
import requests
from functools import wraps
import hashlib
import zipfile
import shutil
import contextlib
import tarfile


def get_data_dir(name=None):
    """ Returns the directories in which dirty_cat looks for data.

    This is typically useful for the end-user to check where the data is
    downloaded and stored.

    """
    # assuming we are in datasets.utils, this calls the module
    module_path = os.path.dirname(os.path.dirname(__file__))
    data_dir = os.path.join(module_path, 'data')
    if name is not None:
        data_dir = os.path.join(data_dir, name)
    return data_dir


@wraps(requests.get)
def request_get(*args, **kwargs):
    return requests.get(*args, **kwargs)


def md5_hash(string):
    m = hashlib.md5()
    m.update(string.encode('utf-8'))
    return m.hexdigest()


def _md5_sum_file(path):
    """ Calculates the MD5 sum of a file.
    """
    with open(path, 'rb') as f:
        m = hashlib.md5()
        while True:
            data = f.read(8192)
            if not data:
                break
            m.update(data)
    return m.hexdigest()


def _check_if_exists(path, remove=False):
    if remove:
        try:
            os.remove(path)
        except OSError:
            pass
        return False
    else:
        return os.path.exists(path)


def _uncompress_file(file_, delete_archive=True):
    """Uncompress files contained in a data_set.


    Parameters
    ----------
    file_: path to file
    delete_archive: whether to delete the compressed file afterwards


    Returns
    -------
    None if everything worked out fine
    ValueError otherwise


    Notes
    -----
    only supports zip and gzip
    """
    sys.stderr.write('Extracting data from %s...' % file_)
    data_dir = os.path.dirname(file_)
    # We first try to see if it is a zip file
    try:
        filename, ext = os.path.splitext(file_)
        with open(file_, "rb") as fd:
            header = fd.read(4)
        processed = False
        if zipfile.is_zipfile(file_):
            z = zipfile.ZipFile(file_)
            z.extractall(path=data_dir)
            z.close()
            if delete_archive:
                os.remove(file_)
            file_ = filename
            processed = True
        elif ext == '.gz' or header.startswith(b'\x1f\x8b'):
            import gzip
            gz = gzip.open(file_)
            if ext == '.tgz':
                filename = filename + '.tar'
            out = open(filename, 'wb')
            shutil.copyfileobj(gz, out, 8192)
            gz.close()
            out.close()
            # If file is .tar.gz, this will be handle in the next case
            if delete_archive:
                os.remove(file_)
            file_ = filename
            processed = True
        if os.path.isfile(file_) and tarfile.is_tarfile(file_):
            with contextlib.closing(tarfile.open(file_, "r")) as tar:
                tar.extractall(path=data_dir)
            if delete_archive:
                os.remove(file_)
            processed = True
        if not processed:
            raise IOError(
                "[Uncompress] unknown archive file format: %s" % file_)

        sys.stderr.write('.. done.\n')
    except Exception as e:
        print('Error uncompressing file: %s' % e)
        raise



import warnings

import numpy as np
from scipy import sparse
from sklearn.cluster import KMeans
from sklearn.feature_extraction.text import CountVectorizer, HashingVectorizer
from sklearn.neighbors import NearestNeighbors
from sklearn.preprocessing import OneHotEncoder
from sklearn.utils import check_random_state

from . import string_distances


def ngram_similarity(X, cats, ngram_range, hashing_dim, dtype=np.float64):
    """
    Similarity encoding for dirty categorical variables:
        Given to arrays of strings, returns the
        similarity encoding matrix of size
        len(X) x len(cats)

    ngram_sim(s_i, s_j) =
        ||min(ci, cj)||_1 / (||ci||_1 + ||cj||_1 - ||min(ci, cj)||_1)
    """
    min_n, max_n = ngram_range
    unq_X = np.unique(X)
    cats = np.array([' %s ' % cat for cat in cats])
    unq_X_ = np.array([' %s ' % x for x in unq_X])
    if not hashing_dim:
        vectorizer = CountVectorizer(analyzer='char',
                                     ngram_range=(min_n, max_n),
                                     dtype=dtype)
        vectorizer.fit(np.concatenate((cats, unq_X_)))
    else:
        vectorizer = HashingVectorizer(analyzer='char',
                                       ngram_range=(min_n, max_n),
                                       n_features=hashing_dim, norm=None,
                                       alternate_sign=False,
                                       dtype=dtype)
        # The hashing vectorizer is stateless. We don't need to fit it on the data
        vectorizer.fit(X)
    count_cats = vectorizer.transform(cats)
    count_X = vectorizer.transform(unq_X_)
    # We don't need the vectorizer anymore, delete it to save memory
    del vectorizer
    sum_cats = np.asarray(count_cats.sum(axis=1))
    SE_dict = {}

    for i, x in enumerate(count_X):
        _, nonzero_idx, nonzero_vals = sparse.find(x)
        samegrams = np.asarray((count_cats[:, nonzero_idx].minimum(nonzero_vals)
                                ).sum(axis=1))
        allgrams = x.sum() + sum_cats - samegrams
        similarity = np.divide(samegrams, allgrams)
        SE_dict[unq_X[i]] = similarity.reshape(-1)
    # We don't need the counts anymore, delete them to save memory
    del count_cats, count_X

    out = np.empty((len(X), similarity.size), dtype=dtype)
    for x, out_row in zip(X, out):
        out_row[:] = SE_dict[x]

    return np.nan_to_num(out)


def get_prototype_frequencies(prototypes):
    """
    Computes the frequencies of the values contained in prototypes
    Reverse sorts the array by the frequency
    Returns a numpy array of the values without their frequencies
    """
    uniques, counts = np.unique(prototypes, return_counts=True)
    sorted_indexes = np.argsort(counts)[::-1]
    return uniques[sorted_indexes], counts[sorted_indexes]


def get_kmeans_prototypes(X, n_prototypes, hashing_dim=128,
                          ngram_range=(3, 3), sparse=False, sample_weight=None, random_state=None):
    """
    Computes prototypes based on:
      - dimensionality reduction (via hashing n-grams)
      - k-means clustering
      - nearest neighbor
    """
    vectorizer = HashingVectorizer(analyzer='char', norm=None,
                                   alternate_sign=False,
                                   ngram_range=ngram_range,
                                   n_features=hashing_dim)
    projected = vectorizer.transform(X)
    if not sparse:
        projected = projected.toarray()
    kmeans = KMeans(n_clusters=n_prototypes, random_state=random_state)
    kmeans.fit(projected, sample_weight=sample_weight)
    centers = kmeans.cluster_centers_
    neighbors = NearestNeighbors()
    neighbors.fit(projected)
    indexes_prototypes = np.unique(neighbors.kneighbors(centers, 1)[-1])
    if indexes_prototypes.shape[0] < n_prototypes:
        warnings.warn('Final number of unique prototypes is lower than ' +
                      'n_prototypes (expected)')
    return np.sort(X[indexes_prototypes])


_VECTORIZED_EDIT_DISTANCES = {
    'levenshtein-ratio': np.vectorize(string_distances.levenshtein_ratio),
    'jaro': np.vectorize(string_distances.jaro),
    'jaro-winkler': np.vectorize(string_distances.jaro_winkler),
}


class SimilarityEncoder(OneHotEncoder):
    """Encode string categorical features as a numeric array.

    The input to this transformer should be an array-like of
    strings.
    The method is based on calculating the morphological similarities
    between the categories.
    The categories can be encoded using one of the implemented string
    similarities: ``similarity='ngram'`` (default), 'levenshtein-ratio',
    'jaro', or 'jaro-winkler'.
    This encoding is an alternative to OneHotEncoder in the case of
    dirty categorical variables.

    Parameters
    ----------
    similarity : str {'ngram', 'levenshtein-ratio', 'jaro', or\
'jaro-winkler'}
        The type of pairwise string similarity to use.

    ngram_range : tuple (min_n, max_n), default=(2, 4)
        Only significant for ``similarity='ngram'``. The range of
        values for the n_gram similarity.

    categories : 'auto', 'k-means', 'most_frequent' or a list of lists/arrays
    of values.
        Categories (unique values) per feature:

        - 'auto' : Determine categories automatically from the training data.
        - list : ``categories[i]`` holds the categories expected in the i-th
          column. The passed categories must be sorted and should not mix
          strings and numeric values.
        - 'most_frequent' : Computes the most frequent values for every
           categorical variable
        - 'k-means' : Computes the K nearest neighbors of K-mean centroids
           in order to choose the prototype categories

        The categories used can be found in the ``categories_`` attribute.
    dtype : number type, default np.float64
        Desired dtype of output.
    handle_unknown : 'error' (default) or 'ignore'
        Whether to raise an error or ignore if a unknown categorical feature is
        present during transform (default is to raise). When this parameter
        is set to 'ignore' and an unknown category is encountered during
        transform, the resulting one-hot encoded columns for this feature
        will be all zeros. In the inverse transform, an unknown category
        will be denoted as None.
    hashing_dim : int type or None.
        If None, the base vectorizer is CountVectorizer, else it's set to
        HashingVectorizer with a number of features equal to `hashing_dim`.
    n_prototypes: number of prototype we want to use.
        Useful when `most_frequent` or `k-means` is used.
        Must be a positive non null integer.
    random_state: either an int used as a seed, a RandomState instance or None.
        Useful when `k-means` strategy is used.

    Attributes
    ----------
    categories_ : list of arrays
        The categories of each feature determined during fitting
        (in order corresponding with output of ``transform``).

    References
    ----------

    For a detailed description of the method, see
    `Similarity encoding for learning with dirty categorical variables
    <https://hal.inria.fr/hal-01806175>`_ by Cerda, Varoquaux, Kegl. 2018
    (accepted for publication at: Machine Learning journal, Springer).


    """

    def __init__(self, similarity='ngram', ngram_range=(2, 4),
                 categories='auto', dtype=np.float64,
                 handle_unknown='ignore', hashing_dim=None, n_prototypes=None, random_state=None):
        super().__init__()
        self.categories = categories
        self.dtype = dtype
        self.handle_unknown = handle_unknown
        self.similarity = similarity
        self.ngram_range = ngram_range
        self.hashing_dim = hashing_dim
        self.n_prototypes = n_prototypes
        self.random_state = random_state

        assert categories in [None, 'auto', 'k-means', 'most_frequent']
        if categories in ['k-means', 'most_frequent'] and (n_prototypes is None or n_prototypes == 0):
            raise ValueError('n_prototypes expected None or a positive non null integer')
        if categories == 'auto' and n_prototypes is not None:
            warnings.warn('n_prototypes parameter ignored with category type \'auto\'')

    def get_most_frequent(self, prototypes):
        """ Get the most frequent category prototypes
        Parameters
        ----------
        prototypes : the list of values for a category variable
        Returns
        -------
        The n_prototypes most frequent values for a category variable
        """
        values, _ = get_prototype_frequencies(prototypes)
        return values[:self.n_prototypes]

    def fit(self, X, y=None):
        """Fit the CategoricalEncoder to X.
        Parameters
        ----------
        X : array-like, shape [n_samples, n_features]
            The data to determine the categories of each feature.
        Returns
        -------
        self
        """
        X = self._check_X(X)
        if self.handle_unknown not in ['error', 'ignore']:
            template = ("handle_unknown should be either 'error' or "
                        "'ignore', got %s")
            raise ValueError(template % self.handle_unknown)

        if ((self.hashing_dim is not None) and
                (not isinstance(self.hashing_dim, int))):
            raise ValueError("value '%r' was specified for hashing_dim, "
                             "which has invalid type, expected None or "
                             "int." % self.hashing_dim)

        if self.categories not in ['auto', 'most_frequent', 'k-means']:
            for cats in self.categories:
                if not np.all(np.sort(cats) == np.array(cats)):
                    raise ValueError("Unsorted categories are not yet "
                                     "supported")

        n_samples, n_features = X.shape
        self.categories_ = list()
        self.random_state_ = check_random_state(self.random_state)

        for i in range(n_features):
            Xi = X[:, i]
            if self.categories == 'auto':
                self.categories_.append(np.unique(Xi))
            elif self.categories == 'most_frequent':
                self.categories_.append(self.get_most_frequent(Xi))
            elif self.categories == 'k-means':
                uniques, count = np.unique(Xi, return_counts=True)
                self.categories_.append(
                    get_kmeans_prototypes(uniques, self.n_prototypes, sample_weight=count,
                                          random_state=self.random_state_))
            else:
                if self.handle_unknown == 'error':
                    valid_mask = np.in1d(Xi, self.categories[i])
                    if not np.all(valid_mask):
                        diff = np.unique(Xi[~valid_mask])
                        msg = ("Found unknown categories {0} in column {1}"
                               " during fit".format(diff, i))
                        raise ValueError(msg)
                self.categories_.append(np.array(self.categories[i], dtype=object))

        return self

    def transform(self, X):
        """Transform X using specified encoding scheme.

        Parameters
        ----------
        X : array-like, shape [n_samples, n_features]
            The data to encode.

        Returns
        -------
        X_new : 2-d array, shape [n_samples, n_features_new]
            Transformed input.

        """
        X = self._check_X(X)

        n_samples, n_features = X.shape

        for i in range(n_features):
            Xi = X[:, i]
            valid_mask = np.in1d(Xi, self.categories_[i])

            if not np.all(valid_mask):
                if self.handle_unknown == 'error':
                    diff = np.unique(X[~valid_mask, i])
                    msg = ("Found unknown categories {0} in column {1}"
                           " during transform".format(diff, i))
                    raise ValueError(msg)

        if self.similarity in ('levenshtein-ratio',
                               'jaro',
                               'jaro-winkler'):
            out = []
            vect = _VECTORIZED_EDIT_DISTANCES[self.similarity]
            for j, cats in enumerate(self.categories_):
                unqX = np.unique(X[:, j])
                encoder_dict = {x: vect(x, cats.reshape(1, -1))
                                for x in unqX}
                encoder = [encoder_dict[x] for x in X[:, j]]
                encoder = np.vstack(encoder)
                out.append(encoder)
            return np.hstack(out)

        elif self.similarity == 'ngram':
            min_n, max_n = self.ngram_range
            out = []
            for j, cats in enumerate(self.categories_):
                encoder = ngram_similarity(X[:, j], cats,
                                           ngram_range=(min_n, max_n),
                                           hashing_dim=self.hashing_dim,
                                           dtype=self.dtype)
                out.append(encoder)
            return np.hstack(out)
        else:
            raise ValueError("Unknown similarity: '%s'" % self.similarity)



#!/usr/bin/env python

import os
from setuptools import setup, find_packages

version_file = os.path.join(
    os.path.dirname(__file__), 'dirty_cat', 'VERSION.txt')
with open(version_file) as fh:
    VERSION = fh.read().strip()

description_file = os.path.join(os.path.dirname(__file__), 'README.rst')
with open(description_file) as fh:
    DESCRIPTION = fh.read()


if __name__ == '__main__':
    setup(name='dirty_cat',
          version=VERSION,
          author='Patricio Cerda',
          author_email='patricio.cerda@inria.fr',
          url='http://dirty-cat.github.io/',
          description=("Machine learning with dirty categories."),
          long_description=DESCRIPTION,
          license='BSD',
          classifiers=[
              'Development Status :: 2 - Pre-Alpha',
              'Environment :: Console',
              'Intended Audience :: Science/Research',
              'License :: OSI Approved :: BSD License',
              'Operating System :: OS Independent',
              'Programming Language :: Python :: 3.5',
              'Programming Language :: Python :: 3.6',
              'Topic :: Scientific/Engineering',
              'Topic :: Software Development :: Libraries',
          ],
          platforms='any',
          packages=find_packages(),
          package_data={'dirty_cat': ['VERSION.txt', 'data/midwest_survey/*.csv']},
          install_requires=['sklearn', 'numpy', 'scipy', 'requests'],
          )



import unittest

import numpy as np

try:
    import Levenshtein
except ImportError:
    Levenshtein = False

from dirty_cat import string_distances


def _random_string_pairs(n_pairs=50):
    rng = np.random.RandomState(0)
    characters = list(map(chr, range(10000)))
    pairs = []
    for n in range(n_pairs):
        s1_len = rng.randint(50)
        s2_len = rng.randint(50)
        s1 = ''.join(np.random.choice(characters, s1_len))
        s2 = ''.join(np.random.choice(characters, s2_len))
        pairs.append((s1, s2))
    return pairs


# TODO: some factorization of what is common for distances;
# check results for same examples on all distances
def _check_levenshtein_example_results(levenshtein_dist):
    assert levenshtein_dist('', '') == 0
    assert levenshtein_dist('', 'abc') == 3
    assert levenshtein_dist('abc', '') == 3
    assert levenshtein_dist('abc', 'abc') == 0
    assert levenshtein_dist('abcd', 'abc') == 1
    assert levenshtein_dist('abc', 'abcd') == 1
    assert levenshtein_dist('xbcd', 'abcd') == 1
    assert levenshtein_dist('axcd', 'abcd') == 1
    assert levenshtein_dist('abxd', 'abcd') == 1
    assert levenshtein_dist('abcx', 'abcd') == 1
    assert levenshtein_dist('axcx', 'abcd') == 2
    assert levenshtein_dist('axcx', 'abcde') == 3


def _check_symmetry(dist_func, *args, **kwargs):
    for (a, b) in _random_string_pairs():
        assert dist_func(
            a, b, *args, **kwargs) == dist_func(b, a, *args, **kwargs)


def _check_levenshtein_distances():
    for levenshtein_dist in [
            string_distances.levenshtein_seq]:
        _check_levenshtein_example_results(levenshtein_dist)
        _check_symmetry(levenshtein_dist)
    for (a, b) in _random_string_pairs():
        assert string_distances.levenshtein_array(
            a, b) == string_distances.levenshtein_seq(a, b)
        assert string_distances.levenshtein_seq(
            a, b) == string_distances.levenshtein(a, b)
        if Levenshtein is not False:
            assert string_distances.levenshtein_seq(
                a, b) == Levenshtein.distance(a, b)


def test_levenshtein_ratio():
    # TODO
    # assert string_distances.levenshtein_ratio('', '') == 1
    # assert string_distances.levenshtein_ratio('', 'abc') == 3
    # assert string_distances.levenshtein_ratio('abc', '') == 3
    # assert string_distances.levenshtein_ratio('abc', 'abc') == 0
    # assert string_distances.levenshtein_ratio('abcd', 'abc') == 1
    # assert string_distances.levenshtein_ratio('abc', 'abcd') == 1
    # assert string_distances.levenshtein_ratio('xbcd', 'abcd') == 1
    # assert string_distances.levenshtein_ratio('axcd', 'abcd') == 1
    # assert string_distances.levenshtein_ratio('abxd', 'abcd') == 1
    # assert string_distances.levenshtein_ratio('abcx', 'abcd') == 1
    # assert string_distances.levenshtein_ratio('axcx', 'abcd') == 2
    # assert string_distances.levenshtein_ratio('axcx', 'abcde') == 3
    pass


# Tests for jaro
def test_jaro():
    # If no character in common: similarity is 0
    assert string_distances.jaro('Brian', 'Jesus') == 0
    assert string_distances.jaro_winkler('Brian', 'Jesus') == 0


def test_identical_strings():
    # Test that if 2 strings are the same, the similarity
    for string1, _ in _random_string_pairs(n_pairs=10):
        assert string_distances.jaro(string1, string1) == 1
        assert string_distances.jaro_winkler(string1, string1) == 1
        assert string_distances.levenshtein_ratio(string1, string1) == 1


def test_compare_implementations():
    # Compare the implementations of python-Levenshtein to our
    # pure-Python implementations
    if Levenshtein is False:
        raise unittest.SkipTest
    for string1, string2 in _random_string_pairs(n_pairs=10):
        assert (string_distances._jaro_winkler(string1, string2,
                    winkler=False)
                == Levenshtein.jaro(string1, string2)
               )
        assert (string_distances._jaro_winkler(string1, string2)
                == Levenshtein.jaro_winkler(string1, string2)
               )
        assert (string_distances.levenshtein_ratio(string1, string2)
                == Levenshtein.ratio(string1, string2)
               )



def test_ngram_similarity():
    # TODO
    # assert ...
    for n in range(1, 4):
        _check_symmetry(string_distances.ngram_similarity, n)
